/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
 *  Main authors:
 *     Ignasi Abio <iabio@nicta.com.au>
 */

/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */


/* This file implements parametric merge networks, used for encoding cardinality constraints. For further details, see "A
 * Parametric Approach for Smaller and Better Encodings of Cardinality Constraints", from Ignasi Abio, Robert Nieuwenhuis,
 * Albert Oliveras and Enric Rodriguez-Carbonell, at the 19th International Conference on Principles and Practice of
 * Constraint Programming (2013). */

#ifndef __MINIZINC_PARAMETRIC_MERGE_NETWORK_hh__
#define __MINIZINC_PARAMETRIC_MERGE_NETWORK_hh__


#include <vector>
#include <cassert>
#include <algorithm>
#include <unordered_map>
//#include <core/Solver.h>

namespace Glucose {

struct uu {
  unsigned a,b;
  uu(unsigned _a, unsigned _b):a(_a),b(_b) {}
  bool operator==(const uu &other) const { return (a == other.a && b == other.b); }
};

struct hash_uu {
  size_t operator()(const uu &ab ) const
  {
    return std::hash<unsigned>()(ab.a) ^ (std::hash<unsigned>()(ab.b)<<1);
  }
};

typedef std::unordered_map<uu,double,hash_uu> HashTableMerge;

class Merge {
private:
  double lambda;
  HashTableMerge costs;
  inline double computeDirectCost(unsigned a, unsigned b);
  inline double computeRecCost(unsigned a, unsigned b);
  inline bool method(unsigned a, unsigned b);
  inline void createDirectNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2, std::vector<Lit>& out);
  void createRecNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2, std::vector<Lit>& out);
public:
  inline Merge (double _lambda);
  //inline Merge () {assert(false); }
  inline double cost (unsigned a, unsigned b);
  inline void createNetwork(Solver* s, const std::vector<Lit>& in1, const std::vector<Lit>& in2, std::vector<Lit>& out);
  inline void createTwoComp(Solver* s, const Lit& i1, const Lit& i2, Lit& o1, Lit& o2);
};

class GetMerge {
public:
  static Merge& Instance(double lambda) {
    static std::unordered_map<double,Merge> singleton;
    if (singleton.find(lambda) != singleton.end())
      return singleton.find(lambda)->second;
    else {
      return singleton.emplace(lambda,Merge(lambda)).first->second;
    }
  }
private:
  GetMerge() {}
  ~GetMerge() {}
  GetMerge(const GetMerge&); // Prevent copy-construction
  GetMerge& operator=(const GetMerge&); // Prevent assignment
};

inline Merge::Merge ( double _lambda ):lambda(_lambda) {
  costs[uu(1,1)] = 2*lambda+3;
}

inline double Merge::cost (unsigned a, unsigned b) {
  if ( a > b ) { unsigned aux = a; a = b; b = aux; }
  if ( a == 0 ) return 0;
  uu ab(a,b);
  HashTableMerge::const_iterator got = costs.find(ab);
  if ( got != costs.end () ) {
    double x = got->second;
    return (x>0?x:-x);
  }
  double directCost = computeDirectCost(a,b);
  double recCost = computeRecCost(a,b);
  if ( directCost <= recCost ) {
    costs[ab] = directCost;
    return directCost;
  } else {
    costs[ab] = - recCost;
    return recCost;
  }
}

inline bool Merge::method(unsigned a, unsigned b) {
  cost(a,b);
  uu ab(a,b);
  return costs[ab] >= 0;
}

inline double Merge::computeDirectCost( unsigned a, unsigned b ) {
  return lambda*(a+b)+a*b+a+b;
}

inline double Merge::computeRecCost( unsigned a, unsigned b ) {
  double cost1 = cost((a+1)/2, (b+1)/2);
  double cost2 = cost((a  )/2, (b  )/2);
  return cost1 + cost2 + lambda*2*((a+b-1)/2)+3*((a+b-1)/2);
}

inline void Merge::createNetwork(Solver* s, const std::vector<Lit>& in1, const std::vector<Lit>& in2, std::vector<Lit>& out) {
  unsigned i1 = in1.size(), i2 = in2.size();
  if ( i1 == 0 ) {
    out = in2;
    return;
  }
  if ( i2 == 0 ) {
    out = in1;
    return;
  }
  if ( method(std::min(i1,i2),std::max(i1,i2)) ) createDirectNetwork(s, in1, i1, in2, i2, out);
  else createRecNetwork(s, in1, i1, in2, i2, out);
}

inline void Merge::createDirectNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2,
                                       std::vector<Lit>& out) {
  for ( unsigned i = 0; i < i1+i2; i++ )
    out.push_back(mkLit(s->newVar(), false));
  for ( unsigned i = 0; i < i1; i++ ) {
    vec<Lit> v(2);
    v[0]=in1[i];
    v[1]=~out[i];
    s->addClause(v);
  }
  for ( unsigned i = 0; i < i2; i++ ) {
    vec<Lit> v(2);
    v[0]=in2[i];
    v[1]=~out[i];
    s->addClause(v);
  }
  for ( unsigned i = 0; i < i1; i++ ) {
    for ( unsigned j = 0; j < i2; j++ ) {
      vec<Lit> v(3);
      v[0]=in1[i];
      v[1]=in2[j];
      v[2]=~out[i+j+1];
      s->addClause(v);
    }
  }
}

inline void Merge::createRecNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2,
                                    std::vector<Lit>& out) {
  if ( !(i1 % 2) && !(i2 % 2) ) {
    std::vector<Lit> in11(i1/2), in12(i1/2), in21(i2/2), in22(i2/2), o1, o2;
    for (unsigned i = 0; i < i1; i+=2) {
      in11[i/2]=in1[i];
      in12[i/2]=in1[i+1];
    }
    for (unsigned i = 0; i < i2; i+=2) {
      in21[i/2]=in2[i];
      in22[i/2]=in2[i+1];
    }
    createNetwork(s, in11, in21, o1);
    createNetwork(s, in12, in22, o2);
    out.resize(i1+i2);
    out[0] = o1[0];
    out[i1+i2-1] = o2.back();
    for (unsigned i = 1; i < o1.size(); i++)
      createTwoComp(s, o2[i-1], o1[i], out[2*i-1], out[2*i]);
  } else if (!(i1%2) && (i2%2)) {
    std::vector<Lit> in11(i1/2), in12(i1/2), in21(i2/2+1), in22(i2/2), o1, o2;
    for (unsigned i = 0; i < i1; i+=2) {
      in11[i/2]=in1[i];
      in12[i/2]=in1[i+1];
    }
    in21[0]=in2[0];
    for (unsigned i = 1; i < i2; i+=2) {
      in22[i/2]=in2[i];
      in21[i/2+1]=in2[i+1];
    }
    createNetwork(s, in11, in21, o1);
    createNetwork(s, in12, in22, o2);
    out.resize(i1+i2);
    out[0] = o1[0];
    for (unsigned i = 1; i < o1.size(); i++)
      createTwoComp(s, o2[i-1], o1[i], out[2*i-1], out[2*i]);
  } else if ((i1%2) && !(i2%2)) {
    std::vector<Lit> in11(i1/2+1), in12(i1/2), in21(i2/2), in22(i2/2), o1, o2;
    in11[0]=in1[0];
    for (unsigned i = 1; i < i1; i+=2) {
      in12[i/2]=in1[i];
      in11[i/2+1]=in1[i+1];
    }
    for (unsigned i = 0; i < i2; i+=2) {
      in21[i/2]=in2[i];
      in22[i/2]=in2[i+1];
    }
    createNetwork(s, in11, in21, o1);
    createNetwork(s, in12, in22, o2);
    out.resize(i1+i2);
    out[0] = o1[0];
    for (unsigned i = 1; i < o1.size(); i++)
      createTwoComp(s, o2[i-1], o1[i], out[2*i-1], out[2*i]);
  } else {
    // Case (iA%2) && (iB%2)
    std::vector<Lit> in11(i1/2+1), in12(i1/2), in21(i2/2+1), in22(i2/2), o1, o2;
    in11[0]=in1[0];
    for (unsigned i = 1; i < i1; i+=2) {
      in12[i/2]=in1[i];
      in11[i/2+1]=in1[i+1];
    }
    in21[0]=in2[0];
    for (unsigned i = 1; i < i2; i+=2) {
      in22[i/2]=in2[i];
      in21[i/2+1]=in2[i+1];
    }
    createNetwork(s, in11, in21, o1);
    createNetwork(s, in12, in22, o2);
    out.resize(i1+i2);
    out[0] = o1[0];
    out[i1+i2-1] = o1.back();
    for (unsigned i = 1; i < o1.size()-1; i++)
      createTwoComp(s, o2[i-1], o1[i], out[2*i-1], out[2*i]);
  }
}

inline void Merge::createTwoComp(Solver* s, const Lit& i1, const Lit& i2, Lit& o1, Lit& o2) {
  o1 = mkLit(s->newVar(), false);
  o2 = mkLit(s->newVar(), false);
  vec<Lit> v(2);
  v[0]=i1;
  v[1]=~o1;
  s->addClause(v);
  v[0]=i2;
  s->addClause(v);
  v.push(i1);
  v[1]=~o2;
  s->addClause(v);
}

};

#endif
