/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
 *  Main authors:
 *     Ignasi Abio <iabio@nicta.com.au>
 */

/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */


/*   This file implements parametric simplified merge networks, used for encoding cardinality constraints. For further 
 *   details, see "A Parametric Approach for Smaller and Better Encodings of Cardinality Constraints", from Ignasi Abio,
 *   Robert Nieuwenhuis, Albert Oliveras and Enric Rodriguez-Carbonell, at the 19th International Conference on
 *   Principles and Practice of Constraint Programming (2013). */

#ifndef __MINIZINC_PARAMETRIC_SIMP_MERGE_NETWORK_hh__
#define __MINIZINC_PARAMETRIC_SIMP_MERGE_NETWORK_hh__

#include <toSAT/CC/ParametricMergeNetworks.hh>

namespace Glucose {

struct uuu {
  unsigned a,b,c;
  uuu(unsigned _a, unsigned _b, unsigned _c):a(_a),b(_b),c(_c) {}
  bool operator==(const uuu &other) const { return (a == other.a && b == other.b && c == other.c); }
};
struct hash_uuu {
  size_t operator()(const uuu &abc ) const
  {
    return (std::hash<unsigned>()(abc.a) ^ (std::hash<unsigned>()(abc.b)<<1) >>1) ^ (std::hash<unsigned>()(abc.c)<<1);
  }
};
typedef std::unordered_map<uuu,double,hash_uuu> HashTableSimpMerge;

class SimpMerge {
 private:
  double lambda;
  HashTableSimpMerge costs;
  inline double computeDirectCost(unsigned a, unsigned b, unsigned c);
  inline double computeRecCost(unsigned a, unsigned b, unsigned c);
  inline bool method(unsigned a, unsigned b, unsigned c);
  inline void createDirectNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2, std::vector<Lit>& out,
				  unsigned o);
  inline void createRecNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2, std::vector<Lit>& out,
			       unsigned o);
 public:
  inline SimpMerge (double _lambda);
  //inline SimpMerge () {assert(false); }
  inline double cost (unsigned a, unsigned b, unsigned c);
  inline void createNetwork(Solver* s, const std::vector<Lit>& in1, const std::vector<Lit>& in2, std::vector<Lit>& out, unsigned o);
};


class GetSimpMerge {
public:
  static SimpMerge& Instance(double& lambda) {
    static std::unordered_map<double,SimpMerge> singleton;
    if (singleton.find(lambda) != singleton.end())
      return singleton.find(lambda)->second;
    else {
      return singleton.emplace(lambda,SimpMerge(lambda)).first->second;
    }
  }
private:
  GetSimpMerge() {}
  ~GetSimpMerge() {}
  GetSimpMerge(const GetSimpMerge&); // Prevent copy-construction
  GetSimpMerge& operator=(const GetSimpMerge&); // Prevent assignment
};



inline SimpMerge::SimpMerge( double _lambda ):lambda(_lambda) {
  costs[uuu(1,1,1)] = lambda+2;
}

inline double SimpMerge::cost (unsigned a, unsigned b, unsigned c) {


  //std::cout << "Cost " << a << " " << b << " " << c << std::endl;

  if ( a > c ) a = c;
  if ( b > c ) b = c;
  if ( a + b <= c ) {
    //std::cout << "Cost " << a << " " << b << " " << c << " = " << GetMerge::Instance(lambda).cost(a,b) << std::endl;;
    return GetMerge::Instance(lambda).cost(a,b);
  }
  uuu abc(a,b,c);
  HashTableSimpMerge::const_iterator got = costs.find(abc);
  if (got != costs.end ()) {
    double x = got->second;
    //std::cout << "Cost " << a << " " << b << " " << c << " = " << (x>0?x:-x) << std::endl;;
    return (x>0?x:-x);
  }
  double directCost = computeDirectCost(a,b,c);
  double recCost = computeRecCost(a,b,c);
  if ( directCost <= recCost ) {
    costs[abc] = directCost;
    //std::cout  << "Cost " << a << " " << b << " " << c << " = " << directCost << std::endl;
    return directCost;
  } else {
    costs[abc] = - recCost;
    //std::cout  << "Cost " << a << " " << b << " " << c << " = " << recCost << std::endl;
    return recCost;
  }
}

inline bool SimpMerge::method(unsigned a, unsigned b, unsigned c) {
  cost(a,b,c);
  uuu abc(a,b,c);
  return costs[abc] >= 0;
}

inline double SimpMerge::computeDirectCost( unsigned a, unsigned b, unsigned c ) {
    return lambda*c+(a+b)*c-c*(c-1)/2-a*(a-1)/2-b*(b-1)/2;
}

inline double SimpMerge::computeRecCost( unsigned a, unsigned b, unsigned c ) {
  double cost1 = cost((a+1)/2, (b+1)/2, c/2+1);
  double cost2 = cost((a  )/2, (b  )/2, c/2  );
  if (c%2)
    return cost1 + cost2 + lambda*(c-1)+(3*c-3)/2;
  else
    return cost1 + cost2 + lambda*(c-1)+(3*c-2)/2;
}

inline void SimpMerge::createNetwork(Solver* s, const std::vector<Lit>& in1, const std::vector<Lit>& in2, std::vector<Lit>& out, unsigned o) {

  //std::cout << "SIMPMERGE " << in1.size() << " " << in2.size() << " " << o << std::endl;

  unsigned i1 = in1.size(), i2 = in2.size();
  if ( i1 > o ) i1 = o;
  if ( i2 > o ) i2 = o;
  if ( i1 + i2 <= o )
    return GetMerge::Instance(lambda).createNetwork (s, in1, in2, out);
  if ( method(std::min(i1,i2),std::max(i1,i2), o) )
    createDirectNetwork(s, in1, i1, in2, i2, out, o);
  else
    createRecNetwork(s, in1, i1, in2, i2, out, o);
}

inline void SimpMerge::createDirectNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2,
					   std::vector<Lit>& out, unsigned o) {
  out.resize(o);
  for ( unsigned i = 0; i < o; i++ ) {
    out[i]=mkLit(s->newVar(), false);
  }
  for ( unsigned i = 0; i < i1; i++ ) {
    vec<Lit> v(2);
    v[0]=in1[i];
    v[1]=~out[i];
    s->addClause(v);
  }
  for ( unsigned i = 0; i < i2; i++ ) {
    vec<Lit> v(2);
    v[0]=in2[i];
    v[1]=~out[i];
    s->addClause(v);
  }
  for ( unsigned i = 0; i < i1; i++ ) {
    for ( unsigned j = 0; i+j+1<o && j < i2; j++ ) {
      vec<Lit> v(3);
      v[0]=in1[i];
      v[1]=in2[j];
      v[2]=~out[i+j+1];
      s->addClause(v);
    }
  }
}

inline void SimpMerge::createRecNetwork(Solver* s, const std::vector<Lit>& in1, unsigned i1, const std::vector<Lit>& in2, unsigned i2,
					std::vector<Lit>& out, unsigned o) {
  if (!(o%2)) {
    std::vector<Lit> in11((i1+1)/2), in12(i1/2), in21((i2+1)/2), in22(i2/2), o1, o2;
    for (unsigned i = 0; i < i1; i++) {
      if (i % 2)
	in12[i/2] = in1[i];
      else
        in11[i/2] = in1[i];
    }
    for (unsigned i = 0; i < i2; i++) {
      if (i % 2)
	in22[i/2] = in2[i];
      else
	in21[i/2] = in2[i];
    }
    createNetwork(s,in11,in21,o1,o/2+1);
    createNetwork(s,in12,in22,o2,o/2);
    out.resize(o);
    out[0] = o1[0];
    for (unsigned i = 1; i < o1.size()-1; i++)
      GetMerge::Instance(lambda).createTwoComp(s,o2[i-1], o1[i], out[2*i-1], out[2*i]);
    out[o-1]=mkLit(s->newVar(), false);
    vec<Lit> v(2);
    v[0]=o1.back();
    v[1]=~out[o-1];
    s->addClause(v);
    v[0]=o2.back();
    s->addClause(v);
  } else {
    std::vector<Lit> in11((i1+1)/2), in12(i1/2), in21((i2+1)/2), in22(i2/2), o1, o2;
    for (unsigned i = 0; i < i1; i++) {
      if (i % 2)
	in12[i/2] = in1[i];
      else
	in11[i/2] = in1[i];
    }
    for (unsigned i = 0; i < i2; i++) {
      if (i % 2)
	in22[i/2] = in2[i];
      else 
        in21[i/2] = in2[i];
    }
    createNetwork(s,in11,in21,o1,o/2+1);
    createNetwork(s,in12,in22,o2,o/2);
    out.resize(o);
    out[0] = o1[0];
    for (unsigned i = 1; i < o1.size(); i++)
      GetMerge::Instance(lambda).createTwoComp(s, o2[i-1], o1[i], out[2*i-1], out[2*i]);
  }
}

};

#endif
