/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
 *  Main authors:
 *     Ignasi Abio <iabio@nicta.com.au>
 */

/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef __MINIZINC_MDD_HH__
#define __MINIZINC_MDD_HH__

//#include <minizinc/ast.hh>
#include <algorithm>
#include <vector>
#include <set>
#include <climits>
#include <toSAT/LI/interval.hh>
//#include <core/Solver.h>


#define NO_NODE (UINT_MAX)
#define TRUE_LIT (UINT_MAX-1)
#define FALSE_LIT (UINT_MAX-2)

namespace Glucose {

class Node {
private:
  std::vector<NodeID> children;
  Interval interval;
  Var v;
  uint level;
public:
  inline Node ( const std::vector<NodeID>& _children, Interval& _interval, uint _level);
  inline Node ( bool b, NodeID id );
  inline Interval getInterval() const;
  inline Var getVar() const;
  inline void setVar( Var _v);
  inline uint getLevel () const;
  inline void setLevel (uint _level);
  inline uint numChildren () const;
  inline NodeID getChild ( uint x ) const;
};

class MDD {
private:
  std::vector<std::set<Interval> > lib;
  std::vector<Node> nodes;
  std::vector<bool> pols;
  std::vector<int> coeffs;
  std::vector<int> domains;
  int K;
  NodeID root;
  NodeID findNode(int currentK, uint level);
public:
  inline MDD(const std::vector<int>& _coeffs, const std::vector<int>& _domains, int _K );
  inline MDD();
  inline void defineConstraint(const std::vector<int>& _coeffs,
			       const std::vector<int>& _domains);
  inline bool isEmpty() const;
  inline NodeID getLastNode() const;
  inline NodeID getFirstNode() const;
  inline bool isLast( NodeID n) const;
  inline NodeID nextNode ( NodeID n ) const;
  inline Var getVar ( NodeID n ) const;
  inline void setVar ( NodeID n, Var v );
  inline uint getLevel ( NodeID n ) const;
  inline bool isTerminal ( NodeID n ) const;
  inline void setTerminal ( NodeID n, bool b );
  inline bool isTrue ( NodeID n ) const;
  inline NodeID getSon ( NodeID n, int numSon ) const;
  inline uint getNumSons ( NodeID n ) const;
  inline uint getSize() const { return nodes.size(); }
  inline NodeID getRoot() const { return root; }
  inline NodeID newRoot(int _K);
};


////////////////////////////////////////////////////////////////////////
////////////////////// Node Class Implementation: //////////////////////
////////////////////////////////////////////////////////////////////////

inline Node::Node ( const std::vector<NodeID>& _children, Interval& _interval, uint _level):
  children(_children),interval(_interval),v(var_Undef),level(_level) {
  //assert ( children.size()>0 && level != TRUE_LIT && level != FALSE_LIT );
}
inline Node::Node ( bool b, NodeID id ):children(0),interval(b,id), v(var_Undef), level(b?TRUE_LIT:FALSE_LIT) {}
inline Interval Node::getInterval() const { return interval; }
inline Var Node::getVar () const { return v; }
inline void Node::setVar(Var _v) { v=_v; }
inline uint Node::getLevel() const { return level; }
inline void Node::setLevel(uint _level) { level=_level; }
inline uint Node::numChildren () const { return children.size(); }
inline NodeID Node::getChild ( uint x ) const { 
  //assert( x < numChildren() );
  return children[x];
}


///////////////////////////////////////////////////////////////////////
////////////////////// MDD Class Implementation: //////////////////////
///////////////////////////////////////////////////////////////////////

inline MDD::MDD(const std::vector<int>& _coeffs, const std::vector<int>& _domains, int _K ):
  lib(_coeffs.size()+1),coeffs(_coeffs),domains(_domains),K(_K) {
  //assert( coeffs.size() == domains.size() );
  root = findNode(K,0);
}
inline MDD::MDD( ) { root = NO_NODE; }
inline void MDD::defineConstraint(const std::vector<int>& _coeffs, const std::vector<int>& _domains) {
  coeffs  = _coeffs;
  domains = _domains;
  lib.resize(coeffs.size()+1);
}
inline bool MDD::isEmpty() const { return root == NO_NODE; }
inline NodeID MDD::getLastNode() const {
  //assert(nodes.size()>0 && root != NO_NODE );
  return (NodeID)nodes.size()-1;
}
inline NodeID MDD::getFirstNode() const {
  //assert(nodes.size()>0 && root != NO_NODE );
  return (NodeID)0;
}
inline bool MDD::isLast ( NodeID n ) const {
  //assert((uint)n<nodes.size());
  return ((uint)n==nodes.size()-1);
}
inline NodeID MDD::nextNode ( NodeID n ) const {
  //assert(!isLast(n));
  return ++n;
}
inline bool MDD::isTerminal ( NodeID n ) const {
  //assert((uint)n<nodes.size());
  uint lvl = nodes[n].getLevel();
  return ( lvl == TRUE_LIT ) || ( lvl == FALSE_LIT );
}
inline void MDD::setTerminal ( NodeID n, bool b ) {
  //assert((uint)n<nodes.size() && !isTerminal(n));
  nodes[n].setLevel(b?TRUE_LIT:FALSE_LIT);
}
inline bool MDD::isTrue ( NodeID n ) const {
  //assert(isTerminal(n));
  return nodes[n].getLevel() == TRUE_LIT;
}
inline NodeID MDD::getSon ( NodeID n, int numSon ) const {
  //assert(numSon < getNumSons(n));
  return nodes[n].getChild(numSon);
}
inline uint MDD::getNumSons ( NodeID n ) const {
  //assert(!isTerminal(n));
  return nodes[n].numChildren();
}
inline Var MDD::getVar ( NodeID n ) const {
  const Node* nn =&(nodes[n]);
  Var v = nn->getVar();
  //assert ( v != var_Undef );
  return v;
}
inline void MDD::setVar( NodeID n, Var v ) {
  Node* nn =&(nodes[n]);
  //assert ( nn->getVar() == var_Undef );
  nn->setVar(v);
}
inline uint MDD::getLevel ( NodeID n ) const {
  //assert(!isTerminal(n));
  return nodes[n].getLevel();
}
inline NodeID MDD::newRoot(int _K) {
  K = _K;
  root = findNode(K,0);
  return root;
}

};

#endif
